##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Dependable framework interfaces
"""

__docformat__ = 'restructuredtext'

from zope.interface import Interface
from zope.exceptions.interfaces import UserError

class IDependable(Interface):
    """Objects that other objects depend on.

    Note that `IDependable` will normally be implemented by an adapter.
    """

    def addDependent(location):
        """Add a dependency to a dependent object by location

        The location is the physical path to the dependent object.
        """
    def removeDependent(location):
        """Remove a dependency with a dependent object by location.

        The location is the physical path to the dependent object.
        """
    def dependents():
        """Return a sequence of dependent object locations.
        """

class DependencyError(UserError):
    """ This object is dependable"""



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Dependable Framework.

"""

__docformat__ = 'restructuredtext'

from zope.interface import implementer
from zope.traversing.api import getParent, canonicalPath, getPath
from zope.annotation.interfaces import IAnnotations

from zope.app.dependable.interfaces import IDependable

class PathSetAnnotation(object):
    """Abstract base class for annotations that are sets of paths.

    To make this into a concrete class, a subclass must set the class
    attribute `key` to a unique annotation key.  A subclass may also
    choose to rename the methods.
    """

    def __init__(self, context):
        self.context = context
        try:
            parent = getParent(self.context)
        except TypeError:
            parent = None

        if parent is not None:
            try:
                pp = getPath(parent)
            except TypeError:
                pp = ""
            if not pp.endswith("/"):
                pp += "/"
            self.pp = pp # parentpath
        else:
            self.pp = ""
        self.pplen = len(self.pp)

    def addPath(self, path):
        path = self._make_relative(path)
        annotations = IAnnotations(self.context)
        old = annotations.get(self.key, ())
        fixed = [self._make_relative(o) for o in old]
        if path not in fixed:
            fixed.append(path)
        new = tuple(fixed)
        if new != old:
            annotations[self.key] = new

    def removePath(self, path):
        path = self._make_relative(path)
        annotations = IAnnotations(self.context)
        old = annotations.get(self.key, ())
        if old:
            fixed = [self._make_relative(o) for o in old]
            fixed = [loc for loc in fixed if loc != path]
            new = tuple(fixed)
            if new != old:
                if new:
                    annotations[self.key] = new
                else:
                    del annotations[self.key]

    def getPaths(self):
        annotations = IAnnotations(self.context)
        locs = annotations.get(self.key, ())
        return tuple(map(self._make_absolute, locs))

    def _make_relative(self, path):
        if path.startswith("/") and self.pp:
            path = canonicalPath(path)
            if path.startswith(self.pp):
                path = path[self.pplen:]
                # Now, the path should not actually begin with a /.
                # canonicalPath doesn't allow trailing / in a path
                # segment, and we already cut off the whole length of
                # the parent, which we guaranteed to begin and end
                # with a /. But it's possible that older dependencies
                # than we test with could produce this scenario, so we
                # leave it for BWC.
                path = path.lstrip("/")

        return path

    def _make_absolute(self, path):
        if not path.startswith("/") and self.pp:
            path = self.pp + path
        return path

@implementer(IDependable)
class Dependable(PathSetAnnotation):
    """See `IDependable`."""

    key = "zope.app.dependable.Dependents"

    addDependent = PathSetAnnotation.addPath
    removeDependent = PathSetAnnotation.removePath
    dependents = PathSetAnnotation.getPaths



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Unit tests for Dependable class.
"""

import unittest

from zope.annotation.attribute import AttributeAnnotations
from zope.location.interfaces import ILocationInfo
from zope.interface import implementer, verify, directlyProvides
from zope.lifecycleevent import ObjectRemovedEvent

from zope.location.interfaces import ILocationInfo

from zope.testing.cleanup import CleanUp

import zope.app.dependable
from zope.app.dependable.dependency import CheckDependency
from zope.app.dependable.interfaces import IDependable, DependencyError


class C(object):
    pass


@implementer(IDependable, ILocationInfo)
class DummyObject(object):

    def dependents(self):
        return ['dependency1', 'dependency2']

    def getPath(self):
        return '/dummy-object'


class TestDependable(unittest.TestCase):

    def factory(self, obj=None):
        from zope.app.dependable import Dependable
        return Dependable(obj if obj is not None else AttributeAnnotations(C()))

    def testVerifyInterface(self):
        object = self.factory()
        verify.verifyObject(IDependable, object)

    def testBasic(self):
        dependable = self.factory()
        self.failIf(dependable.dependents())
        dependable.addDependent('/a/b')
        dependable.addDependent('/c/d')
        dependable.addDependent('/c/e')
        dependable.addDependent('/c/d')
        dependents = list(dependable.dependents())
        dependents.sort()
        self.assertEqual(dependents, ['/a/b', '/c/d', '/c/e'])
        dependable.removeDependent('/c/d')
        dependents = list(dependable.dependents())
        dependents.sort()
        self.assertEqual(dependents, ['/a/b', '/c/e'])
        dependable.removeDependent('/c/d')
        dependents = list(dependable.dependents())
        dependents.sort()
        self.assertEqual(dependents, ['/a/b', '/c/e'])

    def testRelativeAbsolute(self):
        obj = self.factory()
        # Hack the object to have a parent path
        obj.pp = "/a/"
        obj.pplen = len(obj.pp)
        obj.addDependent("foo")
        self.assertEqual(obj.dependents(), ("/a/foo",))
        obj.removeDependent("/a/foo")
        self.assertEqual(obj.dependents(), ())
        obj.addDependent("/a/bar")
        self.assertEqual(obj.dependents(), ("/a/bar",))
        obj.removeDependent("bar")
        self.assertEqual(obj.dependents(), ())

    def testCheckDependency(self):
        obj = DummyObject()
        parent = object()
        event = ObjectRemovedEvent(obj, parent, 'oldName')
        self.assertRaises(DependencyError, CheckDependency, event)

    def testWithParent(self):
        grandparent = C()
        grandparent.__name__ = 'root'

        parent = C()
        parent.__name__ = 'parent'
        parent.__parent__ = grandparent

        obj = AttributeAnnotations(C())
        obj.__parent__ = parent
        obj.__name__ = 'obj'

        from zope.traversing.api import getPath

        # If we can't get the parent path, it's just /
        dependable = self.factory(obj)
        self.assertEqual('/', dependable.pp)

        # If we can, it's whatever it reported, always with a
        # trailing /
        directlyProvides(parent, ILocationInfo)
        parent.getPath = lambda: '/root/parent'
        dependable = self.factory(obj)
        self.assertEqual('/root/parent/', dependable.pp)

        parent.getPath = lambda: '/root/parent/'
        dependable = self.factory(obj)
        self.assertEqual('/root/parent/', dependable.pp)


        dependable.addDependent('/root/parent/sibling/nephew')
        dependents = list(dependable.dependents())
        self.assertEqual(dependents, ['/root/parent/sibling/nephew'])


class TestConfiguration(CleanUp, unittest.TestCase):

    def test_configuration(self):
        from zope.configuration import xmlconfig
        xmlconfig.file('configure.zcml', package=zope.app.dependable)


def test_suite():
    return unittest.defaultTestLoader.loadTestsFromName(__name__)


if __name__=='__main__': # pragma: no cover
    unittest.main(defaultTest='test_suite')
